/*    terminalcontrol.cpp
 *
 * A terminal UI control class
 *
 * Copyright 2002-2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <gdi.h>
#include <eikenv.h>
#include <e32svr.h>
#include "terminalcontrol.h"
#include "s2font.h"

_LIT(KPanic, "TerminalControl");

_LIT(KTerminalFont, "Terminal");
_LIT(KTerminalControl, "terminalcontrol.cpp");

#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KTerminalControl, __LINE__))


CTerminalControl *CTerminalControl::NewL(CS2Font &aFont, const TRect &aRect,
                                         RWindow &aContainerWindow) {
    CTerminalControl *self = new (ELeave) CTerminalControl(aFont);
    CleanupStack::PushL(self);
    self->ConstructL(aRect, aContainerWindow);
    CleanupStack::Pop();
    return self;
}



CTerminalControl::CTerminalControl(CS2Font &aFont) :
    iFont(aFont) {

    TSize size = aFont.FontSize();    
    iFontHeight = size.iHeight;
    iFontWidth = size.iWidth;
}


CTerminalControl::~CTerminalControl() {

    delete iRowBitmap;
    delete [] iChars;
    delete [] iAttributes;
}


void CTerminalControl::ConstructL(const TRect &aRect,
                                  RWindow &aContainerWindow) {

    SetContainerWindowL(aContainerWindow);
    
    SetRect(aRect);
    Resize();
    AllocateBuffersL();
    Clear(KRgbWhite);
}


void CTerminalControl::SetObserver(MTerminalObserver *aObserver) {
    iObserver = aObserver;
}


void CTerminalControl::RemoveObserver() {
    iObserver = NULL;
}



void CTerminalControl::SetGrayed(TBool aGrayed) {
    iGrayed = aGrayed;
    DrawDeferred();
}


// Resizes the terminal in response to size or font change
void CTerminalControl::Resize() {

    TRect rect = Rect();
    TInt width = rect.iBr.iX - rect.iTl.iX;
    TInt height = rect.iBr.iY - rect.iTl.iY;

    assert((iFontWidth > 0) && (iFontHeight > 0));
    assert((width >= iFontWidth) && (height >= iFontHeight));
    iCharWidth = width / iFontWidth;
    iCharHeight = height / iFontHeight;
    if ( iObserver )
        iObserver->TerminalSizeChanged(iCharWidth, iCharHeight);
    TRAPD(error, AllocateBuffersL());
    if ( error != KErrNone ) {
        User::Panic(KPanic, error);
    }
}


// (re)allocate character and attribute buffers
void CTerminalControl::AllocateBuffersL() {
    delete [] iChars;
    iChars = NULL;
    delete [] iAttributes;
    iAttributes = NULL;
    delete iRowBitmap;
    iRowBitmap = NULL;

    iChars = new (ELeave) TText[iCharWidth*iCharHeight];
    iAttributes = new (ELeave) TTerminalAttribute[iCharWidth*iCharHeight];

    iRowBitmap = new CFbsBitmap();
    TDisplayMode displayMode = SystemGc().Device()->DisplayMode();
    User::LeaveIfError(iRowBitmap->Create(TSize(iCharWidth*iFontWidth,
                                                iFontHeight),
                                          displayMode));
    Clear(KRgbWhite);
}


// Clear the buffers
void CTerminalControl::Clear(TRgb aBackground) {
    TText *c = iChars;
    TTerminalAttribute *a = iAttributes;
    TInt num = iCharWidth * iCharHeight;
    while ( num-- ) {
        *c++ = ' ';
        a->iFgColor = KRgbBlack;
        a->iBgColor = aBackground;
        a->iBold = EFalse;
        a->iUnderline = EFalse;
        a++;
    }
}


// Draws text on the terminal window. The coordinates are zero-based
// character coordinates inside the terminal.
void CTerminalControl::DrawText(TInt aX, TInt aY, const TDesC &aText,
                                TBool aBold, TBool aUnderline,
                                TRgb aForeground, TRgb aBackground) {
    // Don't draw text while grayed out
    if ( iGrayed ) {
        return;
    }

    // Check that we are at least partially on screen
    if ( (aX < 0) || (aY < 0) || (aX >= iCharWidth) || (aY >= iCharHeight) ) {
        return;
    }

    // Write the text to our character and attribute buffers
    TInt numChars = aText.Length();
    if ( numChars > (iCharWidth - aX) ) {
        numChars = iCharWidth - aX;
    }
    TText *c = &iChars[aX + aY*iCharWidth];
    TTerminalAttribute *a = &iAttributes[aX + aY * iCharWidth];
    for ( TInt i = 0; i < numChars; i++ ) {
        *c++ = aText[i];
        a->iFgColor = aForeground;
        a->iBgColor = aBackground;
        a->iBold = aBold;
        a->iUnderline = aUnderline;
        a++;
    }
    
    // Actually draw the text on screen
    iFont.RenderText(*iRowBitmap, aText.Left(numChars), 0, 0, aForeground,
                     aBackground);
    CWindowGc &gc = SystemGc();
    gc.Activate(Window());
    TInt x = Rect().iTl.iX;
    TInt y = Rect().iTl.iY;
    gc.BitBlt(TPoint(x + aX*iFontWidth, y + aY*iFontHeight), iRowBitmap,
              TRect(0, 0, numChars*iFontWidth, iFontHeight));
    gc.Deactivate();
    CCoeEnv::Static()->WsSession().Flush();
}


// Clears the terminal with a given background color
void CTerminalControl::ClearTerminal(TRgb aColor) {

    // Clear the buffer
    Clear(aColor);

    // Clear the screen
    CWindowGc &gc = SystemGc();
    gc.Activate(Window());
    gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
    gc.SetBrushColor(aColor);
    gc.SetPenStyle(CGraphicsContext::ENullPen);
    gc.DrawRect(Rect());
    gc.Deactivate();
}


// Returns the current size in characters
TSize CTerminalControl::TerminalSize() {
    return TSize(iCharWidth, iCharHeight);
}


// Calculates a size in characters from a size in pixels
TSize CTerminalControl::TerminalSize(TSize aPixelSize) {
    return TSize((aPixelSize.iWidth + (iFontWidth-1)) / iFontWidth,
                 (aPixelSize.iHeight + (iFontHeight-1)) / iFontHeight);
}


// Calculates a size in pixels from a size in characters
TSize CTerminalControl::PixelSize(TSize aTerminalSize) {
    return TSize(aTerminalSize.iWidth * iFontWidth,
                 aTerminalSize.iHeight * iFontHeight);
}


// CCoeControl::OfferKeyEventL()
TKeyResponse CTerminalControl::OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                              TEventCode aType) {

    if ( aType == EEventKey ) {
        TUint code = aKeyEvent.iCode;

        // Pass keypress to the observer
        if ( iObserver )
            iObserver->KeyPressed((TKeyCode)code, aKeyEvent.iModifiers);
        return EKeyWasConsumed;
    }

    return EKeyWasNotConsumed;
}


// CCoeControl::Draw()
void CTerminalControl::Draw(const TRect & /*aRect*/) const {

    CWindowGc &gc = SystemGc();
    gc.Reset();
    RWsSession ws = CCoeEnv::Static()->WsSession();

    if ( iGrayed ) {
        gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
        gc.SetBrushColor(TRgb(0xcccccc));
        gc.SetPenStyle(CGraphicsContext::ENullPen);
        gc.DrawRect(Rect());
    }
    else {
        TInt x0 = Rect().iTl.iX;
        TInt y0 = Rect().iTl.iY;
        
        // We'll find the longest possible run of identical attributes on
        // each line and render those at one go. We'll then blit the bitmap
        // for that row.
        for ( TInt y = 0; y < iCharHeight; y++ ) {
            TText *lChars = &iChars[y * iCharWidth];
            TTerminalAttribute *lAttribs = &iAttributes[y * iCharWidth];
            TInt x = 0;

            while ( x < iCharWidth ) {
                TTerminalAttribute a0 = lAttribs[x];
                TTerminalAttribute *a = &lAttribs[x];
                TInt num = 0;
                while ( ((x+num) < iCharWidth) &&
                        (a->iFgColor == a0.iFgColor) &&
                        (a->iBgColor == a0.iBgColor) &&
                        (a->iBold == a0.iBold) &&
                        (a->iUnderline == a0.iUnderline) ) {
                    a++;
                    num++;
                }
                iFont.RenderText(*iRowBitmap, TPtrC(&lChars[x], num),
                                 x * iFontWidth, 0, a0.iFgColor, a0.iBgColor);
                x += num;
            }
            gc.BitBlt(TPoint(x0, y0 + y*iFontHeight), iRowBitmap);
            ws.Flush();
        }
    }
}


// CCoeControl::SizeChanged()
void CTerminalControl::SizeChanged() {
    Resize();
}


// CCoeControl::InputCapabilities()
TCoeInputCapabilities CTerminalControl::InputCapabilities() const {
    return TCoeInputCapabilities(TCoeInputCapabilities::EAllText |
                                 TCoeInputCapabilities::ENavigation);
}
